/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */


/*!
    \file   lcmaps_return_poolindex.c
    \brief  This interface returns the poolindex, i.e. leaseid, using LCMAPS
    \author Martijn Steenbakkers for the EU DataGrid.

    If LCMAPS is invoked via this interface it will return the poolindex,
    alternatively named leaseid, to the calling application.
    -# lcmaps_return_poolindex_va
        Returns the poolindex based on a variable argument list
    -# lcmaps_return_poolindex
        Returns the poolindex based on fixed arguments:
                security context, buffer and buffer length
    -# lcmaps_return_poolindex_from_gss_cred
        Returns the poolindex based on fixed arguments:
                gss credential, buffer and buffer length
*/

#ifndef LCMAPS_RETURN_POOLINDEX_C
#define LCMAPS_RETURN_POOLINDEX_C

#include "lcmaps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "globus_gss_assist.h"

#include "lcmaps_log.h"

/* Need this to dig into the context */
#include "gssapi_openssl.h"

/* LCMAPS includes */
#include "lcmaps.h"
#include "lcmaps_return_poolindex.h"

/* Default settings */
#define LCMAPS_POLICY_NAME "acquisition_policy"
#define LCMAPS_LOG_FILE "/var/log/lcmaps.log"
#define LCMAPS_LOG_STRING "poolindex"
#define LCMAPS_BUFSIZ 8192
#define LCMAPS_LOG_BUFSIZ 2048

/* Flags to indicate which type of logging is required */
#define DO_USRLOG           ((unsigned short)0x0001)
#define DO_SYSLOG           ((unsigned short)0x0002)

/******************************************************************************
                       Define module specific variables
******************************************************************************/
static FILE *           lcmaps_log_stream = NULL;  /* stream for LCMAPS logging */
static char *           lcmaps_log_string = NULL;  /* (constant) log string to be prepended to log information */
static char *           lcmaps_log_file = NULL;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/
/*
static int           start_logging();
static int           log_line(int , char * , ...);
static int           log_a_string(int , char * , char *);
static int           end_logging();
*/
static gss_cred_id_t dig_up_gss_cred_from_context(gss_ctx_id_t);
static int           lcmaps_return_poolindex_va(int , ...);

#if 0
/******************************************************************************
Function:       start_logging
Description:    Try to open logfile, if needed.
Parameters:
Returns:        0 succes
                1 failure
******************************************************************************/
static int start_logging()
{
    if (!lcmaps_log_stream)
    {
        if ( (lcmaps_log_file) && ( (lcmaps_log_stream = fopen( lcmaps_log_file, "a")) == NULL) )
        {
            fprintf(stderr, "Cannot open logfile %s: %s\n",
                            lcmaps_log_file, strerror(errno));
            if (do_syslog)
            {
                syslog(LOG_ERR, "Cannot open logfile %s: %m", lcmaps_log_file);
            }
            return 1;
        }
    }

    return 0;
}

/******************************************************************************
Function:       log_line
Description:    Log information to file and or syslog
Parameters:
                prty:    syslog priority (if 0 don't syslog)
                fmt:     string format
Returns:        0 succes
                1 failure
******************************************************************************/
static int
log_line(int prty, char * fmt, ...)
{
    va_list pvar;
    char    buf[LCMAPS_LOG_BUFSIZ];
    int     res;

    va_start(pvar, fmt);
    res = vsnprintf(buf, LCMAPS_LOG_BUFSIZ, fmt, pvar);
    va_end(pvar);
    if ( (res >= LCMAPS_LOG_BUFSIZ) || (res < 0) )
    {
        if (do_syslog)
        {
            syslog(LOG_WARNING, "log_line() warning: log string too long (> %d)\n",
            LCMAPS_LOG_BUFSIZ);
        }
        fprintf(stderr, "log_line() warning: log string too long (> %d)\n",
                LCMAPS_LOG_BUFSIZ);
    }
    if (lcmaps_log_file)
    {
        if (lcmaps_log_stream == NULL)
        {
            if (do_syslog)
            {
                syslog(LOG_ERR, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            }
            fprintf(stderr, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            return 1;
        }
        if (lcmaps_log_string == NULL)
        {
            fprintf(lcmaps_log_stream,"LEVEL %d: %s", prty, buf);
        }
        else
        {
            fprintf(lcmaps_log_stream,"%s %d: %s", lcmaps_log_string, prty, buf);
        }
        fflush(lcmaps_log_stream);
    }
    if (do_syslog && prty)
    {
        syslog(prty, buf);
    }

    return 0;
}

/******************************************************************************
Function:       log_a_string
Description:    Log information to file and or syslog
Parameters:
                prty:       syslog priority (if 0 don't syslog)
                fmt:        string format
                the_string: the string
Returns:        0 succes
                1 failure
******************************************************************************/
static int
log_a_string(int prty, char * fmt, char * the_string)
{
    char    buf[LCMAPS_LOG_BUFSIZ];
    int     res;

    res = snprintf(buf, LCMAPS_LOG_BUFSIZ, fmt, the_string);
    if ( (res >= LCMAPS_LOG_BUFSIZ) || (res < 0) )
    {
        if (do_syslog)
        {
            syslog(LOG_WARNING, "log_line() warning: log string too long (> %d)\n",
            LCMAPS_LOG_BUFSIZ);
        }
        fprintf(stderr, "log_line() warning: log string too long (> %d)\n",
                LCMAPS_LOG_BUFSIZ);
    }
    if (lcmaps_log_file)
    {
        if (lcmaps_log_stream == NULL)
        {
            if (do_syslog)
            {
                syslog(LOG_ERR, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            }
            fprintf(stderr, "log_line() error: cannot open file %s\n", lcmaps_log_file);
            return 1;
        }
        if (lcmaps_log_string == NULL)
        {
            fprintf(lcmaps_log_stream,"LEVEL %d: %s", prty, buf);
        }
        else
        {
            fprintf(lcmaps_log_stream,"%s %d: %s", lcmaps_log_string, prty, buf);
        }
        fflush(lcmaps_log_stream);
    }
    if (do_syslog && prty)
    {
        syslog(prty, buf);
    }

    return 0;
}

/******************************************************************************
Function:       end_logging
Description:    Close logfile
Parameters:
Returns:        0 succes
                1 failure
******************************************************************************/
static int end_logging()
{
    int retval = 0;

    if (lcmaps_log_stream)
    {
        if ( (retval = fclose(lcmaps_log_stream)) )
        {
            return retval;
        }
        else
        {
            lcmaps_log_stream = NULL;
            return 0;
        }
    }
    else
    {
        return 0;
    }
}
#endif

/******************************************************************************
Function:       Dig up the user gss credential from the security context
Description:    Dig up the user gss credential from the security context. Uses
                gss_assist libraries and, in case the delegate credential is empty,
                uses internal gsi libraries as well (hmmmm).
Parameters:
                context_handle:  the security context handle
Returns:        the user credential or GSS_C_NO_CREDENTIAL in case of failure
******************************************************************************/
gss_cred_id_t dig_up_gss_cred_from_context(
    gss_ctx_id_t context_handle
)
{
    gss_name_t                          peer = GSS_C_NO_NAME;
    gss_buffer_desc                     peer_name_buffer = GSS_C_EMPTY_BUFFER;
    OM_uint32                           major_status = 0;
    OM_uint32                           minor_status = 0;
    int                                 rc = -1;
    int                                 initiator = -1;
    gss_cred_id_t                       peer_cred_handle = GSS_C_NO_CREDENTIAL;
    gss_cred_id_t                       deleg_cred_handle = GSS_C_NO_CREDENTIAL;
    gss_cred_id_t                       user_cred_handle = GSS_C_NO_CREDENTIAL;

    rc = globus_module_activate(GLOBUS_GSI_GSS_ASSIST_MODULE);
    if (rc != GLOBUS_SUCCESS)
    {
        /* lcmaps_log (1, "Error activating GLOBUS_GSI_GSS_ASSIST_MODULE\n"); */
        lcmaps_log (1, "Error activating GLOBUS_GSI_GSS_ASSIST_MODULE\n");
        return GSS_C_NO_CREDENTIAL;
    }

    rc = globus_module_activate(GLOBUS_GSI_GSSAPI_MODULE);
    if (rc != GLOBUS_SUCCESS)
    {
        lcmaps_log (1, "Error activating GLOBUS_GSI_GSSAPI_MODULE\n");
        return GSS_C_NO_CREDENTIAL;
    }

    major_status = gss_inquire_context(&minor_status,
                                       context_handle,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       &initiator,
                                       GLOBUS_NULL);

    if(GSS_ERROR(major_status))
    {
        lcmaps_log (1, "GSS failed: Cannot inquire context, Major:%8.8x Minor:%8.8x\n",
                   major_status, minor_status);
        goto dig_up_gss_cred_from_context_error;
    }

    major_status = gss_inquire_context(&minor_status,
                                       context_handle,
                                       initiator ? GLOBUS_NULL : &peer,
                                       initiator ? &peer : GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL);

    if(GSS_ERROR(major_status))
    {
        lcmaps_log (1, "GSS failed: Cannot inquire context, Major:%8.8x Minor:%8.8x\n",
                   major_status, minor_status);
        goto dig_up_gss_cred_from_context_error;
    }
    
    major_status = gss_display_name(&minor_status,
                                    peer,
                                    &peer_name_buffer,
                                    GLOBUS_NULL);
    
    if(GSS_ERROR(major_status))
    {
        lcmaps_log (1, "GSS failed: Cannot display name of peer, Major:%8.8x Minor:%8.8x\n",
                   major_status, minor_status);
        gss_release_name(&minor_status, &peer);
        goto dig_up_gss_cred_from_context_error;
    }

    gss_release_name(&minor_status, &peer);

    lcmaps_log (1, "Found name of peer: %s\n", peer_name_buffer.value);
    /*
     * Sometimes the delegated credential returned by
     * globus_gss_assist_accept_sec_context()/gss_accept_sec_context() is empty.
     * In particular this is true for gatekeeper pings
     * (globusrun -a: "authenticate only", should be "authenticate+authorize only").
     * In this case the gatekeeper should not act on behalf of the user so it does 
     * not receive the delegated proxy.
     * However, LCAS and LCMAPS need a user credential for the authorization chain.
     * This user credential is contained in the context, we have to dig it out.
     * This requires two additional header files: gssapi_openssl.h, globus_gsi_gss_constants.h.
     */

    /*
     * Dig up the delegated credential handle, if it is empty, use the non delegated one
     */
    deleg_cred_handle = (gss_cred_id_t)((gss_ctx_id_desc *)context_handle)->deleg_cred_handle;
    if (deleg_cred_handle == GSS_C_NO_CREDENTIAL)
    {
        char *            peer_subject = NULL;
        gss_name_t        peer_cred_name = GSS_C_NO_NAME;
        gss_buffer_desc   peer_cred_buffer_desc = GSS_C_EMPTY_BUFFER;
        gss_buffer_t      peer_cred_buffer = &peer_cred_buffer_desc; 
        OM_uint32         minor_status2;

        lcmaps_log (1, "Trying to use original user proxy ...");
        peer_cred_handle = (gss_cred_id_t)((gss_ctx_id_desc *)context_handle)->peer_cred_handle;
        /* checking the subject name */
        if ((major_status = gss_inquire_cred(&minor_status,
                                             peer_cred_handle,
                                             &peer_cred_name,
                                             NULL,
                                             NULL,
                                             NULL)) == GSS_S_COMPLETE)
        {
            major_status = gss_display_name(&minor_status,
                                            peer_cred_name,
                                            peer_cred_buffer,
                                            NULL);
            gss_release_name(&minor_status2, &peer_cred_name);
        }
        else
        {
            if (lcmaps_log_stream) 
            {
                globus_gss_assist_display_status(lcmaps_log_stream,
                                                 "gss_inquire_cred() failure",
                                                 major_status,
                                                 minor_status,
                                                 0);
            }
            else
            {
                globus_gss_assist_display_status(stderr,
                                                 "gss_inquire_cred() failure",
                                                 major_status,
                                                 minor_status,
                                                 0);
            }
            lcmaps_log (1, "GSS failed (gss_inquire_cred), Major:%8.8x Minor:%8.8x\n",
                 major_status,minor_status);
            goto dig_up_gss_cred_from_context_error;
        }
        peer_subject = (char *)peer_cred_buffer_desc.value;
        if (peer_name_buffer.value != NULL)
        {
            if (strncmp(peer_name_buffer.value, peer_subject, strlen(peer_name_buffer.value)) != 0)
            {
                lcmaps_log (1, "subject from peer credential differs from client name: (%s != %s)\n",
                           peer_subject, peer_name_buffer.value);
                goto dig_up_gss_cred_from_context_error;
            }
        }
        user_cred_handle = peer_cred_handle;
        if (peer_cred_buffer_desc.value != NULL)
        {
            gss_release_buffer(&minor_status2, peer_cred_buffer);
        }
        peer_subject = NULL;
    }
    else
    {
        lcmaps_log (1, "Trying to use delegated user proxy");
        user_cred_handle = deleg_cred_handle;
    }

    return (user_cred_handle);


dig_up_gss_cred_from_context_error:

    globus_module_deactivate(GLOBUS_GSI_GSSAPI_MODULE);
    globus_module_deactivate(GLOBUS_GSI_GSS_ASSIST_MODULE);

    return GSS_C_NO_CREDENTIAL;
}


/******************************************************************************
Function:   lcmaps_return_poolindex_va
Description:
    If LCMAPS is invoked via this interface it will return the poolindex,
    alternatively named leaseid, to the calling application.
    The arguments of this function are passed as a variable argument list.

Parameters:
    narg:  the number of arguments that follow
    ap:    vararg list, that consists of
        - The GSS Security context established during invocation of the calling
          service. This parameter is of type gss_ctx_id_t.
        - A pointer to a buffer. This buffer will contain the poolindex (NUL
          terminated string) upon successful return.
          This parameter is of type char *.
        - The length of the above mentioned buffer.
          This parameter is of type unsigned int.

Returns:
    0: success
    1: failure
******************************************************************************/
/*!
    \fn lcmaps_return_poolindex_va(
        int        narg,
        va_list    ap
    )
    \brief LCMAPS returns the poolindex
    If LCMAPS is invoked via this interface it will return the poolindex,
    alternatively named leaseid, to the calling application.
    The arguments of this function are passed as a variable argument list.

    \param narg  the number of arguments that follow
    \param ap    va_arg list. This list currently contains:
        -# The GSS Security context established during invocation of the calling
           service. This parameter is of type gss_ctx_id_t.
        -# A pointer to a buffer. This buffer will contain the poolindex (NUL
           terminated string) upon successful return.
           This parameter is of type char *.
        -# The length of the above mentioned buffer.
           This parameter is of type unsigned int.

    \retval 0 success.
    \retval 1 failure.
*/
static int lcmaps_return_poolindex_va(
    int        narg,
    ...
)
{
    va_list                             ap;
    gss_ctx_id_t                        context_handle = GSS_C_NO_CONTEXT;
    char *                              poolindex_buffer = NULL;
    unsigned int                        buffer_length = 0;
    char *                              lcmaps_policy_name = NULL;
    int                                 rc = -1;
    gss_cred_id_t                       user_cred_handle = GSS_C_NO_CREDENTIAL;
    char *                              poolindex = NULL;

    /*
     * Take log file and policy name from environment variables.
     * If not defined take the default ones
     */
    lcmaps_log_file = getenv("LCMAPS_LOG_FILE");
    lcmaps_log_file = (lcmaps_log_file ? lcmaps_log_file : LCMAPS_LOG_FILE );
    lcmaps_policy_name = getenv("LCMAPS_POLICY_NAME");
    lcmaps_policy_name = (lcmaps_policy_name ? lcmaps_policy_name : LCMAPS_POLICY_NAME );

    /*
     * The other environment variables should in principle not be touched,
     * because reasonably defaults are provided by the LCMAPS core.
     * An exception is made for LCMAPS_LOG_STRING, because we want to enforce
     * that it is logged which interface is used.
     * setenv with overwrite=0 only overwrites if the variable is non-existant
     */
    setenv("LCMAPS_LOG_STRING", LCMAPS_LOG_STRING, 0);
    lcmaps_log_string = getenv("LCMAPS_LOG_STRING");

    /* Parse arguments from va_list */
    va_start(ap, narg);
    if (narg == 3)
    {
        context_handle = va_arg(ap, gss_ctx_id_t);
        poolindex_buffer = va_arg(ap, char *);
        buffer_length = va_arg(ap, unsigned int);
    }
    else if (narg == 4)
    {
        context_handle = va_arg(ap, gss_ctx_id_t);
        poolindex_buffer = va_arg(ap, char *);
        buffer_length = va_arg(ap, unsigned int);
        user_cred_handle = va_arg(ap, gss_cred_id_t);
    }
    else
    {
        fprintf(stderr, "%s: The number of arguments (%d) should be in the range %d-%d\n", lcmaps_log_string, narg, 3, 4);
        return 1;
    }
    va_end(ap);

#if 0
    if (start_logging())
    {
        fprintf(stderr, "%s: Cannot start logging\n", lcmaps_log_string);
        return 1;
    }
#endif

    /* First initialize LCMAPS */
    if (lcmaps_init_and_log(lcmaps_log_stream, DO_USRLOG|DO_SYSLOG))
    {
        fprintf(stderr, "%s: LCMAPS initialization failure\n", lcmaps_log_string);
        return 1;
    }

    /* 
     * Retrieve clients (delegated) credential
     * Check if by any chance the gss credential was directly given to us and
     * then use it. If not then plan B: use the security context.
     * Plan B:
     *   First check who initiated the context (should not be us, I think) and
     *   retrieve the name of the peer (=client)
     *   Then dig out the peer's credential from the context structure
     *   (not nice, will globus provide a mechanism for this in the future ?)
     */
    if (user_cred_handle == GSS_C_NO_CREDENTIAL)
    {
        /* plan B */
        user_cred_handle = dig_up_gss_cred_from_context(context_handle);
        if (user_cred_handle == GSS_C_NO_CREDENTIAL)
        {
            lcmaps_log (1, "LCMAPS could not find a valid user gss credential\n");
            goto return_poolindex_error;
        }
    }

    /*
     * Now that we have the credential let's run (and terminate) LCMAPS !
     */
#if ALLOW_EMPTY_CREDENTIALS
    rc = lcmaps_run_and_return_poolindex(
            NULL,
            user_cred_handle,
            NULL,
            &poolindex,
            1,
            &lcmaps_policy_name);
#else
    rc = lcmaps_run_and_return_poolindex(
            user_cred_handle,
            NULL,
            &poolindex,
            1,
            &lcmaps_policy_name);
#endif
    if (rc != 0)
    {
        lcmaps_log (1, "LCMAPS failed to do mapping and return poolindex\n");
        if (lcmaps_term())
        {
            lcmaps_log (1, "LCMAPS termination failure\n");
            goto return_poolindex_error;
        }
        goto return_poolindex_error;
    }
    /* Now copy poolindex into buffer */
    if (poolindex)
    {
        if(strlen(poolindex) + 1 > buffer_length)
        {
            lcmaps_log (1, "Buffer (length=%d) too small for poolindex (length=%d)\n",
                       buffer_length-1, poolindex);
            free(poolindex);
            goto return_poolindex_error;
        }
        else
        {
            strncpy(poolindex_buffer, poolindex, buffer_length);
            lcmaps_log_a_string(0, "lcmaps_return_poolindex(): Returning poolindex %s\n", poolindex);
        }
        free(poolindex);
    }
    else
    {
        lcmaps_log (1, "LCMAPS could not find the poolindex\n");
        goto return_poolindex_error;
    }

    rc = lcmaps_term();
    if (rc)
    {
        lcmaps_log (1, "LCMAPS termination failure\n");
        goto return_poolindex_error;
    }

    return 0;

 return_poolindex_error:

    return 1;
}

/******************************************************************************
Function:   lcmaps_return_poolindex
Description:
    If LCMAPS is invoked via this interface it will return the poolindex,
    alternatively named leaseid, to the calling application.

Parameters:
    context_handle:
        The GSS Security context established during invocation of the calling
        service. This parameter is of type gss_ctx_id_t.
    poolindex_buffer:
        A pointer to a buffer. This buffer will contain the poolindex (NUL
        terminated string) upon successful return.
        This parameter is of type char *.
    buffer_length:
        The length of the above mentioned buffer.
        This parameter is of type unsigned int.

Returns:
    0: success
    1: failure
******************************************************************************/
int lcmaps_return_poolindex(
    gss_ctx_id_t context_handle,
    char *       poolindex_buffer,
    unsigned int buffer_length
)
{
    return (lcmaps_return_poolindex_va(3, context_handle, poolindex_buffer, buffer_length));
}

/******************************************************************************
Function:   lcmaps_return_poolindex_from_gss_cred
Description:
    If LCMAPS is invoked via this interface it will return the poolindex,
    alternatively named leaseid, to the calling application.
    Instead of using the security context as an input it uses a gss credential
    (type gss_cred_id_t).

Parameters:
    credential_handle:
        The GSS credential (of type gss_cred_id_t) of the user.
    poolindex_buffer:
        A pointer to a buffer. This buffer will contain the poolindex (NUL
        terminated string) upon successful return.
        This parameter is of type char *.
    buffer_length:
        The length of the above mentioned buffer.
        This parameter is of type unsigned int.

Returns:
    0: success
    1: failure
******************************************************************************/
int lcmaps_return_poolindex_from_gss_cred(
    gss_cred_id_t credential_handle,
    char *        poolindex_buffer,
    unsigned int  buffer_length
)
{
    return (lcmaps_return_poolindex_va(
                4,
                GSS_C_NO_CONTEXT,
                poolindex_buffer,
                buffer_length,
                credential_handle)
           );
}

#endif /* LCMAPS_RETURN_POOLINDEX_C */

/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/lcmaps/src/lcmaps_return_poolindex.c,v $
    $Date: 2011-03-04 21:55:13 +0100 (Fri, 04 Mar 2011) $
    $Revision: 14868 $
    $Author: dennisvd $
******************************************************************************/
